/** Implements double-ended queue in Java.
 *
 * This follows the Python convention:
 * the right of the deque is the front.
 * 
 * append() adds to the right.
 * appendleft() adds to the left.
 *
 * pop() pops from the right.
 * popleft() pops from the left.
 *
 */
public class Deque {
    Node head;
    Node tail;
    boolean empty;

    /////////////////////////////////////////////////////
    // main method
    //
    public static void main(String[] args) throws EmptyDequeException {

        Deque d = new Deque();

        System.out.println("Populating deque...");
        for(int i=1; i<=10; i++ ) {
            d.append(i);
        }
        System.out.println("Done populating deque.");
        
        System.out.println("Popping from deque...");
        while(!d.isEmpty()){
            System.out.println(d.pop());
        }
        System.out.println("Done popping from deque.");

        System.out.println("Populating second deque...");
        for(int i=1; i<=10; i++ ) {
            d.appendLeft(i);
        }
        System.out.println("Done populating second deque.");
        
        System.out.println("Popping from second deque...");
        while(!d.isEmpty()){
            System.out.println(d.popLeft());
        }
        System.out.println("Done popping from second deque.");

    }
    //
    /////////////////////////////////////////////////////

    public Deque() {
        this(null);
    }

    private Deque(Node n) {
        this.head = n;
        this.tail = n;
        if(n==null) { 
            this.empty = true;
        } else {
            this.empty = false;
        }
    }

    public boolean isEmpty(){
        return this.empty;
    }

    public void append(int data) { 
        // New head of list should point to old head of list 
        Node n = new Node(data, this.head);
        // Break deque's link to old head of list, point to new head of list
        this.head = n;
        // If empty, not empty anymore. Update tail to point at head. 
        if(this.empty) {
            this.tail = this.head;
            this.empty = false;
        }
    }
    
    public void appendLeft(int data) { 
        // New tail of list should point to null
        Node n = new Node(data);
        // Get pointer to old tail (may be null)
        Node old_tail = this.tail;
        // Break deque's link to old tail of list, point to new tail of list
        this.tail = n;
        // If empty, not empty anymore. Update head to point to tail.
        if(this.empty) { 
            this.head = this.tail;
            this.empty = false;
        } else {
            // Old tail pointer is not null, b/c list was not empty. 
            // Update its next pointer.
            old_tail.next = this.tail;
        }
    }

    public int pop() throws EmptyDequeException {
        if(this.empty) {
            throw new EmptyDequeException("Error: Empty deque object! Use the empty() method.");
        }
        // Get pointer to old head of list
        Node n = this.head;
        // Break deque's link to old head of list, point to new head of list
        if(this.head.next != null ) { 
            this.head = n.next;
        } else {
            this.head = null;
            this.tail = null;
            this.empty = true;
        }
        // Return the old head of the list 
        return n.data;
    }

    public int popLeft() throws EmptyDequeException {
        if(this.empty) { 
            throw new EmptyDequeException("Error: Empty deque object! Use the empty() method.");
        }
        // Get pointer to old tail of list
        Node old_tail = this.tail;
        // Get pointer to new tail of list 
        Node runner = this.head;
        if( runner.equals(this.tail) ) {
            this.head = null;
            this.tail = null;
            this.empty = true;
            return runner.data;
        }
        while( runner.next != this.tail ) {
            runner = runner.next;
        }
        // Break deque's link to old tail of list, point to new tail of list
        this.tail = runner;
        this.tail.next = null;
        // Return the old tail of the list
        return old_tail.data;
    }
}

class Node {
    int data;
    Node next;

    public Node(int data) { 
        this(data,null);
    }

    public Node(int data, Node next) { 
        this.data = data;
        this.next = next;
    }
}

class EmptyDequeException extends Exception { 
    public EmptyDequeException(String message){
        super(message);
    }
}

