import java.util.Set;
import java.util.Map;
import java.util.TreeMap;
import java.util.Arrays;
import com.google.common.graph.Network;
import com.google.common.graph.NetworkBuilder;
import com.google.common.graph.ImmutableNetwork;
import com.google.common.graph.MutableNetwork;

//// Or be lazy:
//import java.util.*;
//import com.google.common.graph.*;

// See https://google.github.io/styleguide/javaguide.html

/** This class solves the traveling salesman problem on a graph. */
class TSP {


	public static void main(String[] args) { 
		TSP t = new TSP();
		t.solve();
	}


	/**
	 * This class solves the traveling salesman problem on a graph.
	 * It makes use of Google's Guava library.
	 * This implements a recursive backtracking solution
	 * to search for the shortest path.
	 * 
	 * As is, this class hard-codes an adjacency matrix 
	 * for a network of 5 cities.
	 */

	// The actual graph of cities
	ImmutableNetwork<Node,Edge> graph;
	int graphSize;

	// Storage variables used when searching for a solution 
	String[] route;			// store the route
	double this_distance;	// store the total distance
	double min_distance;	// store the shortest path found so far

	/** Default constructor generates the graph and initializes storage variables */
	public TSP() {
		// Build the graph
		this.graph = buildGraph();
		this.graphSize = this.graph.nodes().size();
		
		// Initialize route variable, shared across recursive method instances
		this.route = new String[this.graphSize];
		
		// Initialize distance variable, shared across recursive method instances
		this.this_distance = 0.0;
		this.min_distance = -1.0; // negative min means uninitialized
	}

	/** This method actually constructs the graph. */
	public ImmutableNetwork<Node,Edge> buildGraph() {
		/** Make a NetworkBuilder object for an undirected network and call build().
         *
         * https://github.com/google/guava/wiki/GraphsExplained#building-graph-instances
		 */

		// MutableNetwork is an interface requiring a type for nodes and a type for edges
		MutableNetwork<Node,Edge> roads = NetworkBuilder.undirected().build();

		// Construct a dictionary of nodes (cities)
		String[] cities = {"A","B","C","D","E"};
		Map<String,Node> all_nodes = new TreeMap<String,Node>();
		for(int i = 0; i<cities.length; i++) {
			Node node = new Node(cities[i]);
			all_nodes.put(cities[i], node);

			// Add the nodes to the network
			roads.addNode(node);
		}
		
		// Construct Edges for roads,
		// and add them to a map
		String[] distances = {"A:B:24","A:C:5","A:D:20","A:E:18","B:C:10","B:D:20","C:D:4","C:E:28","D:E:3"};
		Map<String,Edge> all_edges = new TreeMap<String,Edge>();
		for(int j = 0; j<distances.length; j++) {
			// Parse out (city1):(city2):(distance)
			String[] splitresult = distances[j].split(":");
			String left = splitresult[0];
			String right = splitresult[1];
			String label = left + ":" + right;
			int value = Integer.parseInt(splitresult[2]);

			// Add edges to map
			Edge edge = new Edge(left, right, value);
			all_edges.put(label, edge);
			
			// Add edges to network
			roads.addEdge(all_nodes.get(edge.left), all_nodes.get(edge.right), edge);
		}
		
		// Freeze the network
		ImmutableNetwork<Node,Edge> frozen_roads = ImmutableNetwork.copyOf(roads);
		
		return frozen_roads;
	}
	
	
	/** Public solve method will call the recursive backtracking method to search for solutions on the graph */
	public void solve() {
		/** To solve the traveling salesman problem:
		 * Set up the graph, choose a starting node, then call the recursive backtracking method and pass it the starting node.
		 */

		// We need to pass a starting node to recursive backtracking method
		Node startNode = null;
		
		// Grab a node, any node...
		for( Node n : graph.nodes() ) {
			startNode = n;
			break;
		}
		
		// Visit the first node
		startNode.visit();
		
		// Add first node to the route
		this.route[0] = startNode.label;
		
		// Pass the number of choices made
		int nchoices = 1;
		
		// Recursive backtracking
		explore(startNode, nchoices);
	}
	
	/** Recursive backtracking method: explore possible solutions starting at this node, having made nchoices */
	public void explore(Node node, int nchoices) {
		/**
		 * Solution strategy: recursive backtracking.
		 *
		 * Base case:
		 *		- We've visited as many cities as are on the graph.
		 *		- Check if this is a new solution (distance less than the current minimum).
		 *
		 *	Recursive case:
		 *		- Make a choice (mark node as visited, add city to route).
		 *		- Explore the consequences (recursive call).
		 *		- Unmake the choice (mark node as unvisited, remove city from route).
		 *		- Move on to next choice.
		 */

		if(nchoices == graphSize) {
			// 
			// BASE CASE
			//
			if(this.this_distance < this.min_distance || this.min_distance < 0) {
				// Solution base case:
				// if this_distance < min_distance, this is our new minimum distance
				// if min_distance < 0, this is our first minimium distance
				this.min_distance = this.this_distance;
				printSolution();
			} else {
				printFailure();
			}
			
		} else {
			//
			// RECURSIVE CASE
			// 	
			Set<Node> neighbors = graph.adjacentNodes(node);
			for(Node neighbor : neighbors) {
				if(neighbor.visited == false) {
					
					int distance_btwn = -10000;
					
					for( Edge edge : graph.edgesConnecting(node, neighbor) ) {
						distance_btwn = edge.value;
					}

					// Make a choice
					this.route[nchoices] = neighbor.label;
					neighbor.visit();
					this.this_distance += distance_btwn;
					
					// Explore the consequences
					explore(neighbor,nchoices+1);
					
					// Unmake the choice
					this.route[nchoices] = null;
					neighbor.unvisit();
					this.this_distance -= distance_btwn;
				}
				// Move on to the next choice (continue loop)
			}				
		} // End base/recursive case
	}

	/** Print out solution */
	public void printSolution() {
		System.out.print("@@@@@@@@@@\tNEW SOLUTION\t");
		System.out.println("Route: "+Arrays.toString(this.route)
						  +"\tDistance: "+this.min_distance);
	}
	/** Print out failed path */
	public void printFailure() { }
}

/** Node class to represent cities */
class Node {
	public String label;
	public boolean visited; // Helps us to keep track of where we've been on the graph
	public Node(String name){
		this.label = name;
		this.visited = false;
	}
	public void visit(){
		this.visited = true;
	}
	public void unvisit() {
		this.visited = false;
	}
}

/** Node class to represent roads connecting cities */
class Edge {
	public int value;
	public String left, right; // For convenience in construction process. Not necessary.
	public Edge(String left, String right, int value) {
		this.left = left;
		this.right = right;
		this.value = value;
	}
}
